#include "Display.h"
#include <string.h>

#if DOS_DEBUG
  #include <dpmi.h>
  #include <sys/movedata.h>
  #include <inlines/pc.h>
  #include <sys/farptr.h>
#else
  #include "../Ports/Ports.h"
  #define _farsetsel(sel)
  #define _farnspeekb(offs)     (*((uchar*)0xA0000UL+offs))
  #define _farnspokeb(offs,b)   {*((uchar*)0xA0000UL+offs) = b;}
#endif

#define EXCHANGE(type,a,b)      {type tmp=a; a=b; b=tmp;}
#define SORT2(type,a1,a2)       {if (a1>a2) EXCHANGE(type,a1,a2)}

/*
  Standard VGA palette colors (sorted by color number):

   0 - Black            (2/3 intensity)
   1 - Blue             (2/3 intensity)
   2 - Green            (2/3 intensity)
   3 - Cyan             (2/3 intensity)
   4 - Red              (2/3 intensity)
   5 - Magenta          (2/3 intensity)
   6 - Brown/Orange     (2/3 intensity)
   7 - Bright gray      (2/3 intensity)
   8 - Dark gray        (2/3+1/3 intensity)
   9 - Bright blue      (2/3+1/3 intensity)
  10 - Bright green     (2/3+1/3 intensity)
  11 - Bright cyan      (2/3+1/3 intensity)
  12 - Bright red       (2/3+1/3 intensity)
  13 - Bright magenta   (2/3+1/3 intensity)
  14 - Yellow           (2/3+1/3 intensity)
  15 - White            (2/3+1/3 intensity)

  New VGA palette (sorted by RGB):
  RGB (0 - 0/2 intensity, 1 - 1/2 intensity, 2 - 2/2 intensity (Bright))

  000 - Black
  222 - White

  001 - Blue
  002 - Bright blue

  011 - Cyan
  012 - ?Blue-Cyan
  021 - ?Green-Cyan
  022 - Bright cyan
  122 - ?Bright cyan

  010 - Green
  020 - Bright green

  100 - Red
  200 - Bright red

  110 - Brown/Orange
  120 - ?Green-Yellow
  210 - ?Red-Yellow
  220 - Yellow
  221 - ?Yellow-Blue

  101 - Magenta
  102 - ?Blue-Magenta
  201 - ?Red-Magenta
  202 - Bright magenta
  212 - ?Bright magenta

  111 - ?Gray
  112 - ?Gray-Blue
  121 - ?Gray-Green
  211 - ?Gray-Red
*/

#if 0
// Normal VGA colors
#define CBLACK    COLOR_MAKE(0x00, 0x00, 0x00, 0x00)
#define CBLUE     COLOR_MAKE(0x00, 0x00, 0x00, 0xAA)
#define CGREEN    COLOR_MAKE(0x00, 0x00, 0xAA, 0x00)
#define CCYAN     COLOR_MAKE(0x00, 0x00, 0xAA, 0xAA)
#define CRED      COLOR_MAKE(0x00, 0xAA, 0x00, 0x00)
#define CMAGENTA  COLOR_MAKE(0x00, 0xAA, 0x00, 0xAA)
#define CBROWN    COLOR_MAKE(0x00, 0xAA, 0x55, 0x00)
#define CBGRAY    COLOR_MAKE(0x00, 0xAA, 0xAA, 0xAA)
#define CDGRAY    COLOR_MAKE(0x00, 0x55, 0x55, 0x55)
#define CBBLUE    COLOR_MAKE(0x00, 0x55, 0x55, 0xFF)
#define CBGREEN   COLOR_MAKE(0x00, 0x55, 0xFF, 0x55)
#define CBCYAN    COLOR_MAKE(0x00, 0x55, 0xFF, 0xFF)
#define CBRED     COLOR_MAKE(0x00, 0xFF, 0x55, 0x55)
#define CBMAGENTA COLOR_MAKE(0x00, 0xFF, 0x55, 0xFF)
#define CYELLOW   COLOR_MAKE(0x00, 0xFF, 0xFF, 0x55)
#define CWHITE    COLOR_MAKE(0x00, 0xFF, 0xFF, 0xFF)
#else
// New VGA colors
#define CBLACK    COLOR_MAKE(0x00, 0x00, 0x00, 0x00)
#define CBLUE     COLOR_MAKE(0x00, 0x00, 0x00, 0x80)
#define CGREEN    COLOR_MAKE(0x00, 0x00, 0x80, 0x00)
#define CCYAN     COLOR_MAKE(0x00, 0x00, 0x80, 0x80)
#define CRED      COLOR_MAKE(0x00, 0x80, 0x00, 0x00)
#define CMAGENTA  COLOR_MAKE(0x00, 0x80, 0x00, 0x80)
#define CBROWN    COLOR_MAKE(0x00, 0x80, 0x80, 0x00)
#define CBGRAY    COLOR_MAKE(0x00, 0x80, 0x80, 0x80)
#define CDGRAY    COLOR_MAKE(0x00, 0x40, 0x40, 0x40)
#define CBBLUE    COLOR_MAKE(0x00, 0x00, 0x00, 0xFF)
#define CBGREEN   COLOR_MAKE(0x00, 0x00, 0xFF, 0x00)
#define CBCYAN    COLOR_MAKE(0x00, 0x00, 0xFF, 0xFF)
#define CBRED     COLOR_MAKE(0x00, 0xFF, 0x00, 0x00)
#define CBMAGENTA COLOR_MAKE(0x00, 0xFF, 0x00, 0xFF)
#define CYELLOW   COLOR_MAKE(0x00, 0xFF, 0xFF, 0x00)
#define CWHITE    COLOR_MAKE(0x00, 0xFF, 0xFF, 0xFF)
#endif

// New VGA palette
static const uint32 aNewVGAPalette[16]=
{
  CBLACK   ,
  CBLUE    ,
  CGREEN   ,
  CCYAN    ,
  CRED     ,
  CMAGENTA ,
  CBROWN   ,
  CBGRAY   ,
  CDGRAY   ,
  CBBLUE   ,
  CBGREEN  ,
  CBCYAN   ,
  CBRED    ,
  CBMAGENTA,
  CYELLOW  ,
  CWHITE
};

// Mapping of 27 color indices (3 levels per R, G and B) to
// 16 colors (VGA colors)
static const uint32 a27toColor16[3*3*3]=
{
  CBLACK   ,// 000
  CBLUE    ,// 001
  CBBLUE   ,// 002
  CGREEN   ,// 010
  CCYAN    ,// 011
  CCYAN    ,// 012?
  CBGREEN  ,// 020
  CCYAN    ,// 021?
  CBCYAN   ,// 022
  CRED     ,// 100
  CMAGENTA ,// 101
  CMAGENTA ,// 102?
  CBROWN   ,// 110
  CBGRAY   ,// 111?
  CBGRAY   ,// 112?
  CBROWN   ,// 120?
  CBGRAY   ,// 121?
  CBCYAN   ,// 122?
  CBRED    ,// 200
  CMAGENTA ,// 201?
  CBMAGENTA,// 202
  CBROWN   ,// 210?
  CBGRAY   ,// 211?
  CBMAGENTA,// 212?
  CYELLOW  ,// 220
  CYELLOW  ,// 221?
  CWHITE    // 222
};

// Mapping of 27 color indices (3 levels per R, G and B) to
// 16 color indices (VGA colors indices)
static const uchar a27to16[3*3*3]=
{
   0,//CBLACK   ,// 000
   1,//CBLUE    ,// 001
   9,//CBBLUE   ,// 002
   2,//CGREEN   ,// 010
   3,//CCYAN    ,// 011
   3,//CCYAN    ,// 012?
  10,//CBGREEN  ,// 020
   3,//CCYAN    ,// 021?
  11,//CBCYAN   ,// 022
   4,//CRED     ,// 100
   5,//CMAGENTA ,// 101
   5,//CMAGENTA ,// 102?
   6,//CBROWN   ,// 110
   7,//CBGRAY   ,// 111?
   7,//CBGRAY   ,// 112?
   6,//CBROWN   ,// 120?
   7,//CBGRAY   ,// 121?
  11,//CBCYAN   ,// 122?
  12,//CBRED    ,// 200
   5,//CMAGENTA ,// 201?
  13,//CBMAGENTA,// 202
   6,//CBROWN   ,// 210?
   7,//CBGRAY   ,// 211?
  13,//CBMAGENTA,// 212?
  14,//CYELLOW  ,// 220
  14,//CYELLOW  ,// 221?
  15 //CWHITE    // 222
};

// 4x4 table for ordered dithering to 16-colored VGA display
#define ME(x) {((x*128+7)/15)*0x10001UL, ((x*128+7)/15)*0x100UL}
static const uint32 aVGADitherTable[4][4][2] =
{
  {ME( 0), ME(11), ME( 1), ME(10)},
  {ME(14), ME( 4), ME(15), ME( 5)},
  {ME( 2), ME( 9), ME( 3), ME( 8)},
  {ME(12), ME( 6), ME(13), ME( 7)}
};

// Converts RGB value to VGA color index
inline uchar RGBtoVGAIndex (int x, int y, uint32 color)
{
  uint32 rb, g, r;
  x &= 3;
  y &= 3;
  rb = color & 0xFF00FFUL;
  rb += aVGADitherTable[y][x][0];
  rb = (rb >> 7) & 0x30003UL;
  g  = color & 0x00FF00UL;
  g += aVGADitherTable[y][x][1];
  g >>= 15;
  r = rb >> 16;
  rb &= 3;
  return a27to16[(r*3+g)*3+rb];
}


int Display_Init (tDisplay* pDisplay, const tDisplayInit* pDisplayInit)
{
#if DOS_DEBUG
  __dpmi_regs r;
#else
#endif
  int i, colorsel, modecontrol;

  if
  (!(
      (pDisplayInit->Width == 640) &&
      (pDisplayInit->Height == 480) &&
      (pDisplayInit->ColorDepth == dcd_4BPP)
  ))
  return 0;

  // init gfx mode 640x480x16 colors
#if DOS_DEBUG
  pDisplay->vsel = __dpmi_segment_to_descriptor (0xA000);
  r.x.ax = 0x12;
  __dpmi_int (0x10, &r);
#else
#endif

  // read and save VGA palette
  inportb (0x3da);
  outportb (0x3c0, 0x20|0x14);
  colorsel = inportb (0x3c1);
  inportb (0x3da);
  outportb (0x3c0, 0x20|0x10);
  modecontrol = inportb (0x3c1);
  for (i=0;i<16;i++)
  {
    int j;
    uchar r, g, b;
    inportb (0x3da);
    outportb (0x3c0, 0x20|i);   // 4-bit pixel color
    j = inportb (0x3c1) & 0x3f; // 4/6-bit color index
    if (modecontrol & 0x80)
      j = (j & 0xf) | ((colorsel & 0x3) << 4); // 6-bit color index
    j |= (colorsel & 0xc) << 4; // 8-bit color index
    outportb (0x3c7, j);
    r = inportb (0x3c9) << 2;// read 6-bit R color component
    g = inportb (0x3c9) << 2;// read 6-bit G color component
    b = inportb (0x3c9) << 2;// read 6-bit B color component
    pDisplay->aOldVGAPalette[i] = COLOR_MAKE(0,r,g,b);
  }

  // map pixel colors (0...15) to palette entries (0...15)
  inportb (0x3da);
  for (i=0;i<16;i++)
  {
    outportb (0x3c0, i);        // pixel color #
    outportb (0x3c0, i);        // palette entry #
  }
  outportb (0x3c0, 0x20);

  // set RGB triples for each palette entry in use
  outportb (0x3c8, 0);
  for (i=0;i<16;i++)
  {
    outportb (0x3c9, COLOR_GET_RED  (aNewVGAPalette[i]) >> 2);
    outportb (0x3c9, COLOR_GET_GREEN(aNewVGAPalette[i]) >> 2);
    outportb (0x3c9, COLOR_GET_BLUE (aNewVGAPalette[i]) >> 2);
  }

  pDisplay->Width = 640;
  pDisplay->Height = 480;
  pDisplay->ColorDepth = dcd_4BPP;
  pDisplay->ModeNo = 0x12;

  return 1;
}

void Display_ThreadFunction (tDisplay* pDisplay)
{
}

void Display_Done (tDisplay* pDisplay)
{
#if DOS_DEBUG
  __dpmi_regs r;
#else
#endif
  int i;

  // restore VGA palette
  outportb (0x3c8, 0);
  for (i=0;i<16;i++)
  {
    outportb (0x3c9, COLOR_GET_RED  (pDisplay->aOldVGAPalette[i]) >> 2);
    outportb (0x3c9, COLOR_GET_GREEN(pDisplay->aOldVGAPalette[i]) >> 2);
    outportb (0x3c9, COLOR_GET_BLUE (pDisplay->aOldVGAPalette[i]) >> 2);
  }

#if DOS_DEBUG
  r.x.ax = 0x3;
  __dpmi_int (0x10, &r);
#else
#endif
}

static int FindExactVGAIndex (int x, int y, uint32 color, uchar* pIndex)
{
  uint32 r, g, b, i;
  r = (COLOR_GET_RED  (color) + 64) >> 7;
  g = (COLOR_GET_GREEN(color) + 64) >> 7;
  b = (COLOR_GET_BLUE (color) + 64) >> 7;
  i = (r*3+g)*3+b;
  if (a27toColor16[i] == color)
  {
    *pIndex = a27to16[i];
    return 1;
  }
  else
  {
    *pIndex = RGBtoVGAIndex (x, y, color);
    return 0;
  }
}

inline void PutPixelVGA (uint x, uint y, uchar index)
{
  uint32 offs = (y<<4) + (y<<6) + (x>>3);
  outportw (0x3ce, 8 | (0x8000U>>(x&7)));       // set bit mask
  outportw (0x3ce, 0x205);                      // set 2nd write mode
  _farnspeekb (offs);                           // read latch register
  _farnspokeb (offs, index);                    // set pixel, write latch register
}

inline void SplitScanlineVGA (int xmin, int xmax,
                              uint* pl1, uint* pl2, uint* pl3,
                              uchar* pm1, uchar* pm3)
{
  *pl1 = (8 - (xmin & 7)) & 7;
  *pl3 = (1 + (xmax & 7)) & 7;
  *pl2 = (xmax - xmin + 1) - (*pl1) - (*pl3);
  *pm1 = (*pl1) ? 0xFF << (*pl1) : 0;
  *pm3 = (*pl3) ? 0xFF >> (*pl3) : 0;
  if ((xmin^xmax) < 8)
  {
    *pl1 = xmax - xmin + 1;
    *pl3 = *pl2 = 0;
    *pm1 |= *pm3;
  }
}

void Display_Pixel   (tDisplay* pDisplay, int x, int y, uint32 color)
{
  uchar index;

  if
  (
    (x < 0) ||
    (x >= (int)pDisplay->Width) ||
    (y < 0) ||
    (y >= (int)pDisplay->Height)
  )
  return;

  FindExactVGAIndex (x, y, color, &index);

  _farsetsel (pDisplay->vsel);
  PutPixelVGA (x, y, index);
}

void Display_LineH   (tDisplay* pDisplay,
                      int x, int y, uint len, uint32 color)
{
  uchar index;
  uint l1, l2, l3;
  uchar m1, m3;
  uint32 offs;
  int exact;
  uchar aScanline[4][80];

  if
  (
    (x-1+(int)len < 0) ||
    (x >= (int)pDisplay->Width) ||
    (y < 0) ||
    (y >= (int)pDisplay->Height) ||
    (!len)
  )
  return;

  if (x < 0)
  {
    len -= -x;
    x = 0;
  }
  if ((exact = x + (int)len - (int)pDisplay->Width) > 0)
    len -= exact;

  exact = FindExactVGAIndex (0, 0, color, &index);

  _farsetsel (pDisplay->vsel);
  outportw (0x3ce, 5);
  outportw (0x3ce, 0xFF08U);

  offs = (y<<4) + (y<<6) + (x>>3);
  SplitScanlineVGA (x, x+len-1, &l1, &l2, &l3, &m1, &m3);
  l2 >>= 3;
  l1 = l1 != 0;
  l3 = l3 != 0;

  {
    int plane, l;
    uchar mask = 0x80 >> (x & 7);
#if 01
    for (plane=0; plane<4; plane++)
    {
      outportw (0x3ce, 0x04 | (plane<<8));
      if (l1)
        aScanline[plane][0] = _farnspeekb (offs) & m1;
      if (l2)
        memset (&aScanline[plane][l1], 0, l2);
      if (l3)
        aScanline[plane][l1+l2] = _farnspeekb (offs+l1+l2) & m3;
    }
#else
    if (l1+l2+l3)
      for (plane=0; plane<4; plane++)
        memset (&aScanline[plane][0], 0, l1+l2+l3);
#endif

    for (l=0; len; x++,len--)
    {
      if (!exact)
        index = RGBtoVGAIndex (x, y, color);
      if (index & 1) aScanline[0][l] |= mask;
      if (index & 2) aScanline[1][l] |= mask;
      if (index & 4) aScanline[2][l] |= mask;
      if (index & 8) aScanline[3][l] |= mask;
      mask >>= 1;
      if (!mask)
      {
        mask = 0x80;
        l++;
      }
    }

#if 01
    for (plane=0; plane<4; plane++)
    {
      outportw (0x3c4, 0x02 | (0x100<<plane));
      for (l=0; l<l1+l2+l3; l++)
        _farnspokeb (offs+l, aScanline[plane][l]);
    }
#else
    if (l1)
    {
      l = 0;
      _farnspeekb (offs+l);
      outportw (0x3ce, 8 | (((uint)~m1)<<8));
      for (plane=0; plane<4; plane++)
      {
        outportw (0x3c4, 0x02 | (0x100<<plane));
        _farnspokeb (offs+l, aScanline[plane][l]);
      }
    }

    if (l2)
    {
      outportw (0x3ce, 0xFF08U);
      for (plane=0; plane<4; plane++)
      {
        outportw (0x3c4, 0x02 | (0x100<<plane));
        for (l=l1; l<l1+l2; l++)
          _farnspokeb (offs+l, aScanline[plane][l]);
      }
    }

    if (l3)
    {
      l = l1+l2;
      _farnspeekb (offs+l);
      outportw (0x3ce, 8 | (((uint)~m3)<<8));
      for (plane=0; plane<4; plane++)
      {
        outportw (0x3c4, 0x02 | (0x100<<plane));
        _farnspokeb (offs+l, aScanline[plane][l]);
      }
    }
#endif
  }

  outportw (0x3c4, 0xF02);
}

void Display_LineV   (tDisplay* pDisplay,
                      int x, int y, uint len, uint32 color)
{
  uchar index;
  uint32 offs;
  int exact;

  if
  (
    (x < 0) ||
    (x >= (int)pDisplay->Width) ||
    (y-1+(int)len < 0) ||
    (y >= (int)pDisplay->Height) ||
    (!len)
  )
  return;

  if (y < 0)
  {
    len -= -y;
    y = 0;
  }
  if ((exact = y + len - pDisplay->Height) > 0)
    len -= exact;

  exact = FindExactVGAIndex (0, 0, color, &index);

  offs = (y<<4) + (y<<6) + (x>>3);
  _farsetsel (pDisplay->vsel);
  outportw (0x3ce, 8 | (0x8000U>>(x&7)));       // set bit mask
  outportw (0x3ce, 0x205);                      // set 2nd write mode

  while (len--)
  {
    if (!exact)
      index = RGBtoVGAIndex (x, y++, color);
    _farnspeekb (offs);                           // read latch register
    _farnspokeb (offs, index);                    // set pixel, write latch register
    offs += 80;
  }
}

void Display_Line    (tDisplay* pDisplay,
                      int x1, int y1, int x2, int y2, uint32 color)
{
  int sx, sy, dx1, dy1, dx2, dy2, x, y, m, n, k, cnt;
  uchar index;
  int exact;

  sx = x2 - x1;
  sy = y2 - y1;

  m = ABS(sx);
  n = ABS(sy);

  if (!sy)
    return Display_LineH (pDisplay, MIN(x1,x2), y1, 1 + m, color);
  if (!sx)
    return Display_LineV (pDisplay, x1, MIN(y1,y2), 1 + n, color);

  exact = FindExactVGAIndex (0, 0, color, &index);

  dx2 = dx1 = SGN(sx);
  dy2 = dy1 = SGN(sy);

  if (m < n) {
    int t = m; m = n; n = t;
    dx2 = 0;
  } else {
    dy2 = 0;
  }

  _farsetsel (pDisplay->vsel);

  x = x1;
  y = y1;
  k = m >> 1;
  cnt = m + 1;
  while (cnt--)
  {
    if
    (!(
      (x < 0) ||
      (x >= (int)pDisplay->Width) ||
      (y < 0) ||
      (y >= (int)pDisplay->Height)
    ))
    {
      if (!exact)
        index = RGBtoVGAIndex (x, y, color);
      PutPixelVGA (x, y, index);
    }
    k += n;
    if (k < m) {
      x += dx2;
      y += dy2;
    } else {
      k -= m;
      x += dx1;
      y += dy1;
    }
  }
}

void Display_Box     (tDisplay* pDisplay, int x1, int y1, int x2, int y2,
                      int IsSolid, uint32 color)
{
  int xmin, xmax, ymin, ymax;
  xmin = MIN(x1,x2);
  xmax = MAX(x1,x2);
  ymin = MIN(y1,y2);
  ymax = MAX(y1,y2);
  if (IsSolid)
  {
#if 0
    int y;
    for (y=ymin; y<=ymax; y++)
      Display_LineH (pDisplay, xmin, y, 1+xmax-xmin, color);
#else
    uint l1, l2, l3;
    uchar m1, m3;
    int x, y;
    SplitScanlineVGA (xmin, xmax, &l1, &l2, &l3, &m1, &m3);
    if (l2)
      for (y=ymin; y<=ymax; y++)
        Display_LineH (pDisplay, xmin+l1, y, l2, color);
    if (l1)
      for (x=xmin+l1-1; x>=xmin; x--)
        Display_LineV (pDisplay, x, ymin, 1+ymax-ymin, color);
    if (l3)
      for (x=xmax-l3+1; x<=xmax; x++)
        Display_LineV (pDisplay, x, ymin, 1+ymax-ymin, color);
#endif
  }
  else
  {
    Display_LineH (pDisplay, xmin, ymin, 1+xmax-xmin, color);
    Display_LineH (pDisplay, xmin, ymax, 1+xmax-xmin, color);
    Display_LineV (pDisplay, xmin, ymin, 1+ymax-ymin, color);
    Display_LineV (pDisplay, xmax, ymin, 1+ymax-ymin, color);
  }
}

inline void* AdvancePixelPointer (const void* p, tDisplayColorDepth ColorDepth,
                                  int32 PixelCount)
{
  if (ColorDepth == dcd_15BPP)
    ColorDepth = dcd_16BPP;
  return ((char*)p) + ((PixelCount*ColorDepth)>>3);
}

inline uint32 ReadPixel (const void* p, tDisplayColorDepth ColorDepth)
{
  uint32 c;
  switch (ColorDepth)
  {
    case dcd_32BPP:
      c = *(uint32*)p;
    break;

    case dcd_24BPP:
      c = *((uchar*)p+2);
      c <<= 16;
      c |= *(uint16*)p;
    break;

    case dcd_8BPP:
    case dcd_4BPP:
    default:
      c = 0;
    break;
  }
  return c;
}

void Display_Bitmap  (tDisplay* pDisplay, int x, int y,
                      const void* pBitmap, tDisplayColorDepth ColorDepth,
                      uint BitmapWidth, uint BitmapHeight,
                      uint x1, uint y1, uint x2, uint y2)
{
  int xmin, xmax, ymin, ymax;
  uint l1, l2, l3;
  uchar m1, m3;
  void *p1;
  uint32 offs1;
  uchar aScanline[4][80];

  SORT2 (uint, x1, x2);
  SORT2 (uint, y1, y2);

  xmin = x;
  ymin = y;
  xmax = x + (x2 - x1);
  ymax = y + (y2 - y1);
  if (xmin < 0)
  {
    x1 += -xmin;
    xmin = 0;
  }
  if (ymin < 0)
  {
    y1 += -ymin;
    ymin = 0;
  }
  xmax = MIN(xmax,(int)pDisplay->Width-1);
  ymax = MIN(ymax,(int)pDisplay->Height-1);
  if ((xmin > xmax) || (ymin > ymax))
    return;

  p1 = AdvancePixelPointer (pBitmap, ColorDepth, y1*BitmapWidth + x1);

  _farsetsel (pDisplay->vsel);
  outportw (0x3ce, 5);
  outportw (0x3ce, 0xFF08U);

  offs1 = (ymin<<4) + (ymin<<6) + (xmin>>3);
  SplitScanlineVGA (xmin, xmax, &l1, &l2, &l3, &m1, &m3);
  l2 >>= 3;
  l1 = l1 != 0;
  l3 = l3 != 0;

  for (y=ymin; y<=ymax; y++)
  {
    int plane, l;
    uchar mask = 0x80 >> (xmin & 7);
    uint32 offs2 = offs1;
    void* p2 = p1;
    offs1 += 80;
    p1 = AdvancePixelPointer (p1, ColorDepth, BitmapWidth);

#if 01
    for (plane=0; plane<4; plane++)
    {
      outportw (0x3ce, 0x04 | (plane<<8));
      if (l1)
        aScanline[plane][0] = _farnspeekb (offs2) & m1;
      if (l2)
        memset (&aScanline[plane][l1], 0, l2);
      if (l3)
        aScanline[plane][l1+l2] = _farnspeekb (offs2+l1+l2) & m3;
    }
#else
    if (l1+l2+l3)
      for (plane=0; plane<4; plane++)
        memset (&aScanline[plane][0], 0, l1+l2+l3);
#endif

    for (l=0, x=xmin; x<=xmax; x++)
    {
      uchar index = RGBtoVGAIndex (x, y, ReadPixel(p2,ColorDepth));
      p2 = AdvancePixelPointer (p2, ColorDepth, 1);
      if (index & 1) aScanline[0][l] |= mask;
      if (index & 2) aScanline[1][l] |= mask;
      if (index & 4) aScanline[2][l] |= mask;
      if (index & 8) aScanline[3][l] |= mask;
      mask >>= 1;
      if (!mask)
      {
        mask = 0x80;
        l++;
      }
    }

#if 01
    for (plane=0; plane<4; plane++)
    {
      outportw (0x3c4, 0x02 | (0x100<<plane));
      for (l=0; l<l1+l2+l3; l++)
        _farnspokeb (offs2+l, aScanline[plane][l]);
    }
#else
    if (0*l1)
    {
      l = 0;
      _farnspeekb (offs2+l);
      outportw (0x3ce, 8 | (((uint)~m1)<<8));
      for (plane=0; plane<4; plane++)
      {
        outportw (0x3c4, 0x02 | (0x100<<plane));
        _farnspokeb (offs2+l, aScanline[plane][l]);
      }
    }

    if (l2)
    {
      outportw (0x3ce, 0xFF08U);
      for (plane=0; plane<4; plane++)
      {
        outportw (0x3c4, 0x02 | (0x100<<plane));
        for (l=l1; l<l1+l2; l++)
          _farnspokeb (offs2+l, aScanline[plane][l]);
      }
    }

    if (0*l3)
    {
      l = l1+l2;
      _farnspeekb (offs2+l);
      outportw (0x3ce, 8 | (((uint)~m3)<<8));
      for (plane=0; plane<4; plane++)
      {
        outportw (0x3c4, 0x02 | (0x100<<plane));
        _farnspokeb (offs2+l, aScanline[plane][l]);
      }
    }
#endif
  }

  outportw (0x3c4, 0xF02);
}

void Display_Text    (tDisplay* pDisplay, int x1, int y1, int x2, int y2,
                      int dx1, int dy1, int dx2, int dy2,
                      const char* pText, uint TextStrLen,
                      uint32 BackgroundColor, uint32 TextColor,
                      const uchar pFont8x16[256*16])
{
  uint l1, l2, l3;
  uchar m1, m3;
  uint32 offs1;
  uchar BackColorIndex, TextColorIndex;
  int BackColorExact, TextColorExact;
  int x, y;
  uchar aScanline[4][80];

  SORT2 (int, x1, x2);
  SORT2 (int, y1, y2);
  SORT2 (int, dx1, dx2);
  SORT2 (int, dy1, dy2);
  if
  (
    (dx1 >= (int)pDisplay->Width) ||
    (dy1 >= (int)pDisplay->Height) ||
    (dx2 < 0) ||
    (dy2 < 0)
  )
  return;
  dx1 = MAX (dx1, 0);
  dy1 = MAX (dy1, 0);
  dx2 = MIN (dx2, (int)pDisplay->Width-1);
  dy2 = MIN (dy2, (int)pDisplay->Height-1);

  BackColorExact = FindExactVGAIndex (0, 0, BackgroundColor, &BackColorIndex);
  TextColorExact = FindExactVGAIndex (0, 0, TextColor, &TextColorIndex);

  _farsetsel (pDisplay->vsel);
  outportw (0x3ce, 5);
  outportw (0x3ce, 0xFF08U);

  offs1 = (dy1<<4) + (dy1<<6) + (dx1>>3);
  SplitScanlineVGA (dx1, dx2, &l1, &l2, &l3, &m1, &m3);
  l2 >>= 3;
  l1 = l1 != 0;
  l3 = l3 != 0;

  for (y=dy1; y<=dy2; y++)
  {
    int plane, l;
    uchar mask = 0x80 >> (dx1 & 7);
    uint32 offs2 = offs1;
    offs1 += 80;

    for (plane=0; plane<4; plane++)
    {
      outportw (0x3ce, 0x04 | (plane<<8));
      if (l1)
        aScanline[plane][0] = _farnspeekb (offs2) & m1;
      if (l2)
        memset (&aScanline[plane][l1], 0, l2);
      if (l3)
        aScanline[plane][l1+l2] = _farnspeekb (offs2+l1+l2) & m3;
    }

    for (l=0, x=dx1; x<=dx2; x++)
    {
      uchar index;
      uint i = x - x1;
      uint j = y - y1;
      if
      (
        (i < TextStrLen*8) &&
        (j < 16)
      )
      {
        uint ch = (uchar)pText[i>>3];
        if (pFont8x16[ch*16+j] & (0x80>>(i&7)))
        {
          if (TextColorExact)
            index = TextColorIndex;
          else
            index = RGBtoVGAIndex (x, y, TextColor);
        }
        else
        {
          if (BackColorExact)
            index = BackColorIndex;
          else
            index = RGBtoVGAIndex (x, y, BackgroundColor);
        }
      }
      else
      {
          if (BackColorExact)
            index = BackColorIndex;
          else
            index = RGBtoVGAIndex (x, y, BackgroundColor);
      }
      if (index & 1) aScanline[0][l] |= mask;
      if (index & 2) aScanline[1][l] |= mask;
      if (index & 4) aScanline[2][l] |= mask;
      if (index & 8) aScanline[3][l] |= mask;
      mask >>= 1;
      if (!mask)
      {
        mask = 0x80;
        l++;
      }
    }

    for (plane=0; plane<4; plane++)
    {
      outportw (0x3c4, 0x02 | (0x100<<plane));
      for (l=0; l<l1+l2+l3; l++)
        _farnspokeb (offs2+l, aScanline[plane][l]);
    }
  }

  outportw (0x3c4, 0xF02);
}

#if DOS_DEBUG & 0
#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <stdlib.h>
uchar aBMP[768][1024][3];

void LoadBMP1024x768x24bpp (char *name)
{
  uchar BMPHeader[54];
  FILE *f;
  int x, y;
  f = fopen (name, "rb");
  if (f == NULL)
    return;
  fread (&BMPHeader, sizeof(BMPHeader), 1, f);
  memset (aBMP, 0, sizeof(aBMP));
  for (y=0;y<768;y++)
    for (x=0;x<1024;x++)
    {
      uchar rgb[3];
      fread (&rgb, sizeof(rgb), 1, f);
      // account for upside-down
      aBMP[768-1-y][x][0] = rgb[0]; // B
      aBMP[768-1-y][x][1] = rgb[1]; // G
      aBMP[768-1-y][x][2] = rgb[2]; // R
    }
  fclose (f);
}

int main (int argc, char *argv[])
{
  int i, j;
  tDisplay Display;
  tDisplayInit DisplayInit = {640, 480, dcd_4BPP};
  if (!Display_Init (&Display, &DisplayInit))
    return 0;
  if (argc == 2)
    LoadBMP1024x768x24bpp (argv[1]);

  Display_Box (&Display, 0, 0, 639, 479, 1, 0x010101);
  for (j=0; j<4; j++)
    for (i=0; i<4; i++)
      Display_Box (&Display, i*2*640/7, j*2*480/7, (i*2+1)*640/7-1, (j*2+1)*480/7-1,
                   1, aNewVGAPalette[j*4+i]);
  for (j=0; j<4; j++)
    for (i=0; i<4; i++)
      Display_Box (&Display, i*2*640/7, j*2*480/7, (i*2+1)*640/7-1, (j*2+1)*480/7-1,
                   0, aNewVGAPalette[(j*4+i)^7]);
  getch();

  for (j=0; j<480/2; j++)
    for (i=0; i<640/2; i++)
    {
      Display_Pixel (&Display,       i,       j,
                     COLOR_MAKE(0,0xFF*i/(640/2),0,0));
      Display_Pixel (&Display, 640/2+i,       j,
                     COLOR_MAKE(0,0,0xFF*i/(640/2),0));
      Display_Pixel (&Display,       i, 480/2+j,
                     COLOR_MAKE(0,0,0,0xFF*i/(640/2)));
      Display_Pixel (&Display, 640/2+i, 480/2+j,
                     COLOR_MAKE(0,0xFF*i/(640/2),0xFF*i/(640/2),0xFF*i/(640/2)));
    }

  for (j=0; j<480/4; j++)
    for (i=0; i<640/4; i++)
    {
      Display_Pixel (&Display, 640/4*1+i, 480/4*1+j,
                     COLOR_MAKE(0,0xFF*i/(640/4),0xFF*i/(640/4),0x00));
      Display_Pixel (&Display, 640/4*2+i, 480/4*1+j,
                     COLOR_MAKE(0,0x00,0xFF*i/(640/4),0xFF*i/(640/4)));
      Display_Pixel (&Display, 640/4*1+i, 480/4*2+j,
                     COLOR_MAKE(0,0xFF*i/(640/4),0x00,0xFF*i/(640/4)));
    }

  Display_Line (&Display,   0,   0, 639,   0, CBRED);
  Display_Line (&Display, 639,   0, 639, 479, CBGREEN);
  Display_Line (&Display, 639, 479,   0, 479, CBBLUE);
  Display_Line (&Display,   0, 479,   0,   0, CYELLOW);
  Display_Line (&Display,   0,   0, 639, 479, CBCYAN);
  Display_Line (&Display, 639,   0,   0, 479, CBMAGENTA);
  getch();

  i = 0;
  if (argc == 2)
  {
//    while (!kbhit())
    {
      Display_Bitmap (&Display, 0+1, 0+1, (void*)aBMP, dcd_24BPP, 1024, 768, 0+1, 0+1, 639-1, 479-1);
//      Display_Bitmap (&Display, 0+1, 0+1, (void*)aBMP, dcd_24BPP, 1024, 768, 1023-639+1, 767-479+1, 1023-1, 767-1);
      i+=2;
    }
//    while (kbhit())
      getch();
  }

  Display_Done (&Display);
  printf ("%d\n", i);
  return 0;
}
#endif

